import pandas as pd


class Ensembler:
    """
    This class ensembles forecast_models' forecasts to provide a unique
    forecast
    """

    @staticmethod
    def average_top_models(
            df_predictions: pd.DataFrame,
            df_models_ranked: pd.DataFrame,
            n_tops: int
    ) -> pd.DataFrame:
        """
        Average top forecast_models
        """
        # -- Get top forecast_models for each id
        df_models_ranked = (
            df_models_ranked
            .loc[df_models_ranked["ranking"] <= n_tops]
        )
        # -- Filter unwanted forecast_models
        df_predictions = df_predictions.merge(
            df_models_ranked,
            on=["model"],
            how='left'
        )
        df_predictions = (
            df_predictions.loc[~df_predictions['ranking'].isna()]
            .copy()
        )
        # -- Average
        df_predictions.sort_values(["model"], inplace=True)
        df_predictions = df_predictions.groupby(['date']).agg(
            {"forecast": "mean",
             "forecast_lower": "mean",
             "forecast_upper": "mean",
             "model": '/'.join}
        ).reset_index()
        forecast_columns = ["forecast", "forecast_lower", "forecast_upper"]
        df_predictions = df_predictions[forecast_columns + ["date", "model"]]
        return df_predictions

    @staticmethod
    def get_best_models(
            df_predictions: pd.DataFrame,
            df_models_ranked: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Get best model
        """
        # -- Get top forecast_models for each id
        df_models_ranked = (
            df_models_ranked
            .loc[df_models_ranked["ranking"] == 1]
        )
        # -- Filter unwanted forecast_models
        df_predictions = df_predictions.merge(
            df_models_ranked,
            on=["model"],
            how='left'
        )
        df_predictions = (
            df_predictions
            .loc[~df_predictions['ranking'].isna()].copy()
        )
        forecast_columns = ["forecast", "forecast_lower", "forecast_upper"]
        df_predictions = df_predictions[forecast_columns + ["date", "model"]]
        return df_predictions



from forecastout.forecast_engine.model_trainer import ModelTrainer
from forecastout.forecast_engine.model_predictor import ModelPredictor
from forecastout.forecast_engine.backtester import Backtester
from forecastout.forecast_engine.ensemble_models import ensemble_models
from forecastout.forecast_engine.anomaly_fixer import AnomalyFixer
from forecastout.forecast_engine.decompose_time_series \
    import create_df_ts_decomposition


# -- ML main packages
from abc import ABC, abstractmethod
import pandas as pd

class ForecastModel(ABC):

    def __init__(
            self,
            df_train_y:pd.DataFrame,
            dict_config:dict
    ):
        self.df_train_y = df_train_y
        self.dict_config = dict_config

    @abstractmethod
    def do_forecast(self, list_dates: list):
        pass

    @abstractmethod
    def get_feature_importance(self):
        pass



from forecastout.forecast_engine.ensembler import Ensembler
import pandas as pd


def ensemble_models(
        df_predictions: pd.DataFrame,
        df_models_ranked: pd.DataFrame,
        ensemble_method: str,
        n_tops: int
) -> pd.DataFrame:
    """Select the specified algorithm"""
    if ensemble_method == "get_best_models":
        df_ensemble = Ensembler.get_best_models(
            df_predictions=df_predictions,
            df_models_ranked=df_models_ranked
        )
        return df_ensemble
    if ensemble_method == "average_top_models":
        df_ensemble = Ensembler.average_top_models(
            df_predictions=df_predictions.copy(),
            df_models_ranked=df_models_ranked.copy(),
            n_tops=n_tops
        )
        return df_ensemble



import pandas as pd
from typing import List
from forecastout.forecast_engine.forecast_model_factory \
    import ForecastModelFactory


class ModelTrainer:
    def __init__(
            self,
            model_names: List[str],
            df_train: pd.DataFrame,
            config: dict
    ):
        self.model_names = model_names
        self.df_train = df_train
        self.config = config

    @staticmethod
    def filter_config_by(config: dict, model) -> dict:
        return config['models'][model]

    def train(self) -> pd.DataFrame:
        list_output = []
        # --  Training:
        for model_name in self.model_names:
            dict_output_id = {}
            model = ForecastModelFactory.get_model(
                model=model_name,
                df_train_y=self.df_train['value'],
                dict_config=ModelTrainer.filter_config_by(
                    self.config,
                    model_name
                ),
                series_train_dates=self.df_train['date']
            )
            dict_output_id['model'] = model
            list_output.append(dict_output_id)
        return pd.DataFrame(list_output)



import pandas as pd
from forecastout.disaggregation_features.feature_creator \
    import FeatureCreator
from forecastout.disaggregation_features.feature_normalizer \
    import FeatureNormalizer
from forecastout.disaggregation_features.feature_encoding \
    import FeatureEncoder


def add_features(
        df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Create list of features
    """
    # -- Create features
    feature_creator = FeatureCreator(df=df)
    feature_creator.get_day()
    feature_creator.get_month()
    feature_creator.get_year()
    feature_creator.get_day_of_week()
    feature_creator.get_holidays()
    # -- Normalize value
    df = FeatureNormalizer.normalization_base100(
        df=feature_creator.df,
        norm_level=["month", "year"],
        norm_feats=["value"]
    )
    # -- Create more features
    feature_creator = FeatureCreator(df=df)
    feature_creator.get_lags(
        features=["value"],
        number_lags=7
    )
    feature_creator.get_moving_average(
        features=['value'],
        windows=[7, 30]
    )
    df = feature_creator.df.copy()
    # -- Encode features
    feature_encoder = FeatureEncoder()
    df = feature_encoder.numerical_encode(
        df=df.copy(),
        encode_feats=['day', 'day_of_week', 'month'],
        encode_value='value'
    )
    return df



import pandas as pd
from typing import List

class FeatureNormalizer:
    @staticmethod
    def normalization_base100(
            df: pd.DataFrame,
            norm_level: List[str],
            norm_feats: List[str]
    ) -> pd.DataFrame:
        """
        Base 100 normalization by norm_level
        """
        for norm_feat in norm_feats:
            df['max'] = df.groupby(norm_level)[norm_feat].transform('max')
            col_norm = norm_feat #+ '_norm'
            df[col_norm] = (df[norm_feat] / df['max']) * 100
            df.drop('max', axis=1, inplace=True)
        return df



import pandas as pd
from typing import List
from forecastout.data_engine.data_handler import DataHandler


class InputChecker:
    def __init__(
            self,
            df: pd.DataFrame,
            sum_aggregation: bool = True,
            horizon: int = None,
            months_to_backtest: int = None,
            models_to_use: List[str] = None,
            average_top_models_number: int = None
    ):
        # -- Main inputs
        self.df = df
        self.sum_aggregation = sum_aggregation
        self.horizon = horizon
        self.months_to_backtest = months_to_backtest
        self.models_to_use = models_to_use
        self.average_top_models_number = average_top_models_number
        # -- Preliminary Checks
        self.__check_df_instance()
        self.__check_df_column_names()
        self.__check_sum_aggregation()
        self.__check_horizon()
        self.__check_months_to_backtest()
        self.__check_models_to_use()
        self.__check_average_top_models_number()
        self.__check_backtesting_consistency()

    def __check_df_instance(self):
        if not isinstance(self.df, pd.DataFrame):
            error_string = (
                    "Input df should be a pandas DataFrame."
            )
            raise TypeError(error_string)

    def __check_df_column_names(self):
        column_names_list = ['date', 'value']
        for column in self.df.columns:
            if column not in column_names_list:
                error_string = (
                    f"Column name '{column}' for the input " +
                    "df is not acceptable. The acceptable " +
                    "column names for a DataFrame to be " +
                    f"accepted are: {column_names_list}."
                )
                raise ValueError(error_string)

    def __check_backtesting_consistency(self):
        # -- Checks of DF
        data_handler = DataHandler(
            df=self.df,
            horizon=self.horizon,
            sum_aggregation=self.sum_aggregation
        )
        df_monthly = data_handler.df_monthly.copy()
        # -- df_monthly
        num_of_monthly_observations = df_monthly.dropna()['date'].count()
        if num_of_monthly_observations - self.months_to_backtest < 24:
            error_string = (
                "Your data consists in  " +
                f"'{num_of_monthly_observations}' monthly observations, " +
                "while the input parameter months_to_backtest " +
                f"is set to '{self.months_to_backtest}'. " +
                "Make sure that the number of monthly observations " +
                "minus the input parameter months_to_backtest " +
                "is larger than 24."
            )

            raise ValueError(error_string)

    def __check_sum_aggregation(self):
        if not isinstance(self.sum_aggregation, bool):
            error_string = (
                "Input parameter sum_aggregation is " +
                f"{type(self.sum_aggregation)}. " +
                "Expected input type is bool. " +
                "If the df input presents daily data, you can aggregate it " +
                "monthly by adding it in a monthly basis if input " +
                "sum_aggregation is set to True " +
                "or average it if set to False."
            )
            raise TypeError(error_string)

    def __check_horizon(self):
        if not isinstance(self.horizon, int):
            error_string = (
                "Input parameter horizon is " +
                f"{type(self.horizon)}. " +
                "Expected input type is int."
            )
            raise TypeError(error_string)

    def __check_months_to_backtest(self):
        if not isinstance(self.months_to_backtest, int):
            error_string = (
                "Input parameter months_to_backtest is " +
                f"{type(self.months_to_backtest)}. " +
                "Expected input type is int."
            )
            raise TypeError(error_string)

    def __check_models_to_use(self):
        model_names_list = [
            'autoarima',
            'holtwinters',
            'prophet',
            'seasonalnaive'
        ]
        if not isinstance(self.models_to_use, list):
            error_string = (
                    "Input parameter models_to_use is " +
                    f"{type(self.models_to_use)}. " +
                    "Expected input type is List[str]."
            )
            raise TypeError(error_string)
        for model in self.models_to_use:
            if not isinstance(model, str):
                error_string = (
                    "Element of list models_to_use is a " +
                    f"{type(self.models_to_use)}. " +
                    "Expected elements of the list must be all str."
                )
                raise TypeError(error_string)
            if model not in model_names_list:
                error_string = (
                    f"Model name '{model}' is not acceptable. " +
                    f"The acceptable model names are: " +
                    f"{model_names_list}."
                )
                raise ValueError(error_string)

    def __check_average_top_models_number(self):
        if not isinstance(self.average_top_models_number, int):
            error_string = (
                "Input parameter average_top_models_number is "
                f"{type(self.average_top_models_number)}. " +
                "Expected input type is int."
            )
            raise TypeError(error_string)



import pandas as pd
from statsmodels.tsa.seasonal import STL


def create_df_ts_decomposition(
        df: pd.DataFrame
) -> pd.DataFrame:
    """
    Create DF with stl decomposition
    """
    df_hist = df[df['test'] == 0].copy()
    df_hist = pd.Series(
        df_hist['value'].to_list(),
        index=pd.date_range(
            start=df_hist['date'].min(),
            end=df_hist['date'].max(),
            freq="MS"),
        name="ts"
    )
    stl = STL(df_hist, period=12)
    res = stl.fit()
    df_ts_decomposition = pd.DataFrame(
        dict(
            s1=res.seasonal,
            s2=res.trend,
            s3=res.resid
        )
    ).reset_index()
    df_ts_decomposition.columns = ['date', 'seasonal', 'trend', 'residual']
    return df_ts_decomposition



import pandas as pd


class ModelPredictor:
    def __init__(
            self,
            df_models_trained:
            pd.DataFrame,
            df: pd.DataFrame,
            config: dict
    ):
        self.df_models_trained = df_models_trained
        self.df = df
        self.config = config

    def predict(self) -> pd.DataFrame:
        list_output = []
        for model in self.df_models_trained['model'].unique():
            # -- Get test
            dates_test = self.df['date'].tolist()
            # -- Do forecast
            df_forecast_i = model.do_forecast(
                list_dates=dates_test
            )
            # -- Append
            list_output.append(df_forecast_i)
        # -- Convert to df
        df_output = pd.concat(list_output)
        # -- Ensure positivity
        df_num = df_output._get_numeric_data()
        df_num[df_num < 0] = 0
        return df_output



import pandas as pd


def disaggregate_monthly_to_daily_forecast(
        df_daily_shares: pd.DataFrame,
        df_monthly_forecast: pd.DataFrame
) -> pd.DataFrame:
    df_daily_shares["year_month"] = df_daily_shares["date"].astype(
        str).str[0:7]
    df_monthly_forecast["year_month"] = df_monthly_forecast[
                                            "date"].astype(str).str[
                                        0:7]
    df_monthly_forecast.drop('date', axis=1, inplace=True)
    df_daily_forecast = df_daily_shares.merge(
        df_monthly_forecast,
        on='year_month',
        how='left'
    )
    forecast_columns = ["forecast", "forecast_lower", "forecast_upper"]
    for forecast_column in forecast_columns:
        df_daily_forecast[forecast_column] = df_daily_forecast[
                                                 forecast_column] * \
                                             df_daily_forecast[
                                                 "daily_share"]
    relevant_columns = ["date"] + forecast_columns + ["model"]
    df_daily_forecast = df_daily_forecast[relevant_columns].copy()
    return df_daily_forecast



from forecastout.disaggregation_models.abstract_model \
    import DisaggregationModel
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV, RepeatedKFold
import pandas as pd


class RandomForestModel(DisaggregationModel):
    """
    Random Forest Model
    """
    def __init__(self,
                 *args,
                 **kwargs
                 ):
        super(RandomForestModel, self).__init__(*args, **kwargs)
        # -- grid search
        grid = GridSearchCV(
            estimator=RandomForestRegressor(random_state=123),
            param_grid=self.dict_config['param_grid'],
            scoring=self.dict_config['scoring'],
            n_jobs=1,
            cv=RepeatedKFold(
                n_splits=self.dict_config['n_splits'],
                n_repeats=self.dict_config['n_repeats'],
                random_state=123
            ),
            refit=self.dict_config['refit'],
            verbose=0,
            return_train_score=self.dict_config['return_train_score']
        )
        grid.fit(X=self.df_train_x, y=self.df_train_y.values.ravel())
        self.random_forest_model = grid.best_estimator_

    def do_prediction(
            self, 
            list_dates: list, 
            df_test_x: pd.DataFrame) -> pd.DataFrame:
        list_prediction = self.random_forest_model.predict(df_test_x).tolist()
        df_prediction = pd.concat(
            [
                pd.DataFrame(list_dates, columns=['date']),
                pd.DataFrame(list_prediction, columns=['forecast']),
            ],
            axis=1,
        )
        df_prediction['model'] = 'random_forest'
        df_prediction = df_prediction[['forecast', 'model', 'date']].copy()
        return df_prediction

    def get_feature_importance(self) -> pd.DataFrame:
        list_feature_importance = self.random_forest_model.feature_importances_
        df_feature_importance = pd.concat(
            [pd.DataFrame(self.df_train_x.columns,
                          columns=['feature']),
             pd.DataFrame(list_feature_importance,
                          columns=['feature_importance'])
             ], axis=1
        )
        df_feature_importance['model'] = 'random_forest'
        df_feature_importance.sort_values(
            'feature_importance',
            ascending=False,
            inplace=True
        )
        return df_feature_importance



# -- ML main packages
from abc import ABC, abstractmethod
import pandas as pd


class DisaggregationModel(ABC):

    def __init__(
            self,
            df_train_y: pd.DataFrame,
            df_train_x: pd.DataFrame,
            dict_config: dict
    ):
        self.df_train_y = df_train_y.copy()
        self.df_train_x = df_train_x.copy()
        self.dict_config = dict_config

    @abstractmethod
    def do_prediction(self, list_dates: list, df_test_x: pd.DataFrame):
        pass

    @abstractmethod
    def get_feature_importance(self):
        pass



from forecastout.forecast_models.abstract_model import ForecastModel
from pmdarima.arima import auto_arima
import pandas as pd
import numpy as np


class AutoArimaModel(ForecastModel):
    """
    Auto arima model
    """
    def __init__(self,
                 *args,
                 **kwargs
                 ):
        super(AutoArimaModel, self).__init__(*args, **kwargs)
        self.arima_model = (
            auto_arima(self.df_train_y,
                       seasonal=self.dict_config["seasonality"],
                       m=self.dict_config["freq"]
                       )
        )

    def do_forecast(self, list_dates: list) -> pd.DataFrame:
        arima_prediction, arima_conf_int = self.arima_model.predict(
            len(list_dates),
            return_conf_int=True,
            alpha=self.dict_config["alpha_intervals"]
        )
        arima_prediction = pd.concat(
            [
                pd.DataFrame(
                    np.array(arima_prediction.reset_index()[0]),
                    columns=['forecast']),
                pd.DataFrame(
                    arima_conf_int,
                    columns=['forecast_lower', 'forecast_upper'])
            ],
            axis=1)
        arima_prediction['model'] = 'autoarima'
        arima_prediction['date'] = list_dates
        # -- Ensure date
        arima_prediction['date'] = pd.to_datetime(arima_prediction['date'])
        return arima_prediction

    def get_feature_importance(self):
        return None



import pandas as pd
from forecastout.disaggregation_models.random_forest_model \
    import RandomForestModel


class DisaggregationModelFactory:
    @staticmethod
    def get_model(
            disaggregation_model: str,
            df_train_y: pd.DataFrame,
            df_train_x: pd.DataFrame,
            dict_config: dict
    ):
        if disaggregation_model == "random_forest":
            return RandomForestModel(
                df_train_y=df_train_y,
                df_train_x=df_train_x,
                dict_config=dict_config
            )



from forecastout.forecast_models.abstract_model import ForecastModel
from statsforecast.models import SeasonalNaive
from statsforecast import StatsForecast
import pandas as pd


class NaiveSeasonalModel(ForecastModel):
    def __init__(self,
                 series_train_dates: pd.Series,
                 *args,
                 **kwargs
                 ):
        """
        This implementation a naive seasonal forecast
        """
        super(NaiveSeasonalModel, self).__init__(*args, **kwargs)
        # -- Constructor
        # -- Initialize Holtwinters Exponential Smoothing with additive
        # and multiplicative methods
        self.seasonalnaive_model = StatsForecast(models=[
            SeasonalNaive(season_length=self.dict_config['freq'])],
            freq='MS',
            n_jobs=-1
        )
        # -- Transform data
        df_train_y_seasonalnaive = (
            pd.concat([series_train_dates, self.df_train_y], axis=1)
        )
        df_train_y_seasonalnaive.columns = ['ds', 'y']
        df_train_y_seasonalnaive['ds'] = (
            pd.to_datetime(df_train_y_seasonalnaive['ds'])
        )
        df_train_y_seasonalnaive['ds'] = (
            df_train_y_seasonalnaive['ds'].dt.tz_localize(None)
        )
        df_train_y_seasonalnaive['unique_id'] = 'id'
        # -- Fit forecast_models
        self.seasonalnaive_model.fit(df_train_y_seasonalnaive)

    def do_forecast(self, list_dates: list) -> pd.DataFrame:
        # -- Predict
        seasonalnaive_prediction = self.seasonalnaive_model.predict(
            h=len(list_dates),
            level=[self.dict_config['alpha_intervals']]
        ).reset_index()
        # -- Transform data for correct output
        seasonalnaive_prediction = (
            seasonalnaive_prediction.drop(['unique_id'], axis=1)
        )
        seasonalnaive_prediction.columns = [
            'date', 'forecast', 'forecast_lower', 'forecast_upper'
        ]
        seasonalnaive_prediction['model'] = 'seasonalnaive'
        seasonalnaive_prediction['date'] = list_dates
        # -- Ensure date
        seasonalnaive_prediction['date'] = (
            pd.to_datetime(seasonalnaive_prediction['date'])
        )
        return seasonalnaive_prediction

    def get_feature_importance(self):
        return None



import pandas as pd
from forecastout.disaggregation_features.feature_creator \
    import FeatureCreator


class DisaggregationModelPredictor:
    def __init__(self, model, df: pd.DataFrame, config: dict):
        self.model = model
        self.df = df
        self.config = config

    def predict(self) -> pd.DataFrame:
        list_output = []
        df_copy = self.df.copy()
        list_dates_test = df_copy.loc[df_copy['test'] == 1, "date"].tolist()
        for day in list_dates_test:
            # -- Get test
            df_test_x_day = (
                df_copy.loc[
                    df_copy['date'] == day,
                    self.config['disaggregation_features']
                ].copy()
            )
            # -- Do forecast
            df_prediction_i = self.model.do_prediction(
                df_test_x=df_test_x_day,
                list_dates=[day])
            # -- append results
            list_output.append(df_prediction_i)
            # -- UPDATE FEATURES
            df_copy.loc[
                df_copy['date'] == day, "value"
            ] = df_prediction_i['forecast'].values[0]
            feature_creator = FeatureCreator(df=df_copy)
            feature_creator.get_lags(
                features=["value"],
                number_lags=7
            )
            feature_creator.get_moving_average(
                features=['value'],
                windows=[7, 30]
            )
            df_copy = feature_creator.df.copy()
        # -- Convert to df
        df_output = pd.concat(list_output)
        # -- Ensure positivity
        df_num = df_output._get_numeric_data()
        df_num[df_num < 0] = 0
        return df_output



# Main forecasting forecast_models
# -- Python main packages
import pandas as pd

from forecastout.forecast_models.autoarima_model import AutoArimaModel
from forecastout.forecast_models.prophet_model import ProphetModel
from forecastout.forecast_models.holtwinters_model import HoltWintersModel
from forecastout.forecast_models.naive_seasonal import NaiveSeasonalModel


class ForecastModelFactory:
    @staticmethod
    def get_model(
            model,
            df_train_y: pd.DataFrame,
            dict_config: dict,
            series_train_dates: pd.Series
    ):
        if model == "autoarima":
            return AutoArimaModel(
                df_train_y=df_train_y,
                dict_config=dict_config
            )
        if model == "prophet":
            return ProphetModel(
                df_train_y=df_train_y,
                dict_config=dict_config,
                series_train_dates=series_train_dates
            )
        if model == "holtwinters":
            return HoltWintersModel(
                df_train_y=df_train_y,
                dict_config=dict_config,
                series_train_dates=series_train_dates
            )
        if model == "seasonalnaive":
            return NaiveSeasonalModel(
                df_train_y=df_train_y,
                dict_config=dict_config,
                series_train_dates=series_train_dates
            )



import pandas as pd
import numpy as np
from dateutil.relativedelta import relativedelta


class TrainTestSplit:
    """
    This class identifies which observations are for training and for testing
    """

    @staticmethod
    def split_by_current_closed_month(
            df: pd.DataFrame,
            current_month: str,
            granularity_month: bool = True
    ) -> pd.DataFrame:
        min_date = (
            pd.to_datetime(current_month)
            .replace(day=1, hour=0, minute=0, second=0)
        )
        if granularity_month:
            df['test'] = np.where(df["date"] <= min_date, 0, 1)
        else:
            df['test'] = np.where(df["date"] < min_date, 0, 1)
        return df

    @staticmethod
    def split_for_ts_backtest(
            df: pd.DataFrame,
            current_month: str,
            months_to_backtest: int
    ) -> pd.DataFrame:
        # -- Filter data before current month
        current_month = (
            pd.to_datetime(current_month)
            .replace(day=1, hour=0, minute=0, second=0)
        )
        df = df[df["date"] <= current_month].copy()
        # -- Get train/test data
        min_test_date = (
                current_month - relativedelta(months=months_to_backtest)
        )
        df['test'] = np.where(df["date"] <= min_test_date, 0, 1)
        # -- Return
        return df



from forecastout.data_engine import DataHandler, TrainTestSplit
from forecastout.forecast_engine import (
    ModelTrainer,
    ModelPredictor,
    Backtester,
    ensemble_models,
    AnomalyFixer,
    create_df_ts_decomposition
)
from forecastout.disaggregation_features.add_features \
    import add_features
from forecastout.disaggregation_engine import (
    DisaggregationModelTrainer,
    DisaggregationModelPredictor,
    get_daily_shares,
    disaggregate_monthly_to_daily_forecast,
    remake_monthly_forecast_current_month
)
from forecastout.utils import update_config, InputChecker
from logging import Logger
import pandas as pd
import os.path as op
import yaml
import warnings
warnings.filterwarnings("ignore")


class ForecastOut:
    def __init__(
            self,
            df: pd.DataFrame,
            sum_aggregation: bool = True,
            horizon: int = 3,
            months_to_backtest: int = 3,
            models_to_use: list = None,
            average_top_models_number: int = 1
    ):
        # -- Initialize logs
        self.log = Logger(name="Initialization")
        # -- Initialize input
        self.df = df
        self.sum_aggregation = sum_aggregation
        self.horizon = horizon
        self.months_to_backtest = months_to_backtest
        self.models_to_use = models_to_use
        if self.models_to_use is None:
            self.models_to_use = ['autoarima', 'holtwinters', 'prophet']
        self.average_top_models_number = average_top_models_number
        InputChecker(
            df=self.df,
            sum_aggregation=self.sum_aggregation,
            horizon=self.horizon,
            months_to_backtest=self.months_to_backtest,
            models_to_use=self.models_to_use,
            average_top_models_number=self.average_top_models_number
        )
        # -- Initialize output
        self.df_monthly_forecast = pd.DataFrame()
        self.df_predictions_by_model = pd.DataFrame()
        self.df_ts_decomposition = pd.DataFrame()
        self.df_predictions_bt = pd.DataFrame()
        self.df_daily_shares = pd.DataFrame()
        self.df_daily_forecast = pd.DataFrame()
        # -- Read Config
        current_path = op.dirname(__file__)
        with open(op.join(current_path, "config.yaml"), 'rb') as fp:
            self.config = yaml.safe_load(fp)
        self.log.info(">> 0. Update Config")
        self.config = update_config(
            config=self.config,
            horizon=self.horizon,
            months_to_backtest=self.months_to_backtest,
            models_to_use=self.models_to_use,
            average_top_models_number=self.average_top_models_number
        )
        self.log.info(">> 1. Data Treatment")
        data_handler = DataHandler(
            df=self.df,
            horizon=self.config['horizon'],
            sum_aggregation=self.sum_aggregation
        )
        self.df_monthly = data_handler.df_monthly.copy()
        self.df_daily = data_handler.df_daily.copy()
        self.log.info(">> 2. Monthly Forecast")
        self.__make_monthly_forecast()
        self.log.info(">> 3. Daily Disaggregation")
        if data_handler.granularity_of_df == 'daily':
            self.__get_daily_disaggregation_shares()
            self.__make_daily_forecast()
            self.__remake_monthly_forecast()

    def __make_monthly_forecast(self):
        self.log.info(">> 2.1. Train/Test Split")
        df_train_test = TrainTestSplit.split_by_current_closed_month(
            df=self.df_monthly.copy(),
            current_month=(
                self.df_monthly.loc[
                    ~self.df_monthly['value'].isna(), 'date'
                ].max()
            ),
            granularity_month=True
        )
        self.log.info(">> 2.2. Train Models")
        model_trainer = ModelTrainer(
            model_names=self.config["models_to_use"],
            df_train=df_train_test[df_train_test['test'] == 0].copy(),
            config=self.config
        )
        df_models_trained = model_trainer.train()
        self.log.info(">> 2.3. Predict Models")
        model_predictor = ModelPredictor(
            df_models_trained=df_models_trained,
            df=df_train_test[df_train_test['test'] == 1].copy(),
            config=self.config,
        )
        self.df_predictions_by_model = model_predictor.predict()
        self.log.info(">> 2.4. Backtest Models")
        backtester = Backtester(
            df=self.df_monthly.copy(),
            model_names=self.config["models_to_use"],
            config=self.config
        )
        self.df_predictions_bt = backtester.return_predictions()
        df_models_ranked = backtester.return_models_ranked()
        self.log.info(">> 2.5. Ensemble Models")
        df_ensemble = ensemble_models(
            df_predictions=self.df_predictions_by_model.copy(),
            df_models_ranked=df_models_ranked.copy(),
            ensemble_method=self.config["ensemble"]["ensemble_method"],
            n_tops=self.config["ensemble"]["average_top_models"]["n_tops"]
        )
        self.log.info(">> 2.6. Correct anomaly forecast")
        anomaly_fixer = AnomalyFixer(
            df_forecast=df_ensemble.copy(),
            df_actual=df_train_test.copy(),
            config=self.config
        )
        self.df_monthly_forecast = anomaly_fixer.df_forecast.copy()
        self.log.info(">> 2.7. Time Series Decomposition")
        self.df_ts_decomposition = create_df_ts_decomposition(
            df=df_train_test.copy()
        )
        self.log.info(">> 2.X. End of forecast monthly")

    def __get_daily_disaggregation_shares(self):
        self.log.info(">> 3.1. Feature Creation")
        df = add_features(df=self.df_daily)
        self.log.info(">> 3.2. Train/Test Split")
        df_train_test = TrainTestSplit.split_by_current_closed_month(
            df=df.copy(),
            current_month=(
                self.df_monthly.loc[
                    ~self.df_monthly['value'].isna(), 'date'
                ].max()
            ),
            granularity_month=False
        )
        self.log.info(">> 3.3. Train disaggregation models")
        disaggregation_model_trainer = DisaggregationModelTrainer(
            df_train=df_train_test.loc[df_train_test['test'] == 0].copy(),
            config=self.config
        )
        disaggregation_model_trained = disaggregation_model_trainer.train(
            self.config["disaggregation_model_selected"]
        )
        self.log.info(">> 3.4. Predict disaggregation models")
        disaggregation_model_predictor = DisaggregationModelPredictor(
            model=disaggregation_model_trained,
            df=df_train_test.copy(),
            config=self.config
        )
        df_daily_prediction = disaggregation_model_predictor.predict()
        self.log.info(">> 3.5. Get daily shares")
        self.df_daily_shares = get_daily_shares(
            df_daily_prediction=df_daily_prediction.copy(),
            sum_aggregation=self.sum_aggregation
        )

    def __make_daily_forecast(self):
        df_daily_forecast = disaggregate_monthly_to_daily_forecast(
            df_daily_shares=self.df_daily_shares.copy(),
            df_monthly_forecast=self.df_monthly_forecast.copy()
        )
        self.df_daily_forecast = (
            df_daily_forecast.loc[
                df_daily_forecast["date"] > self.df["date"].max()
                ].copy()
        )

    def __remake_monthly_forecast(self):
        self.df_monthly_forecast = remake_monthly_forecast_current_month(
            df_daily_forecast=self.df_daily_forecast.copy(),
            df_actuals=self.df.copy()
        )



import pandas as pd
from forecastout.disaggregation_engine.disaggregation_model_factory \
    import DisaggregationModelFactory


class DisaggregationModelTrainer:
    def __init__(
            self,
            df_train: pd.DataFrame,
            config: dict):
        self.df_train = df_train
        self.config = config

    @staticmethod
    def filter_config_by(config: dict, disaggregation_model: str):
        return config['disaggregation_models'][disaggregation_model]

    def train(self, disaggregation_model: str):
        model = DisaggregationModelFactory.get_model(
            disaggregation_model=disaggregation_model,
            df_train_y=self.df_train['value'],
            df_train_x=self.df_train[self.config["disaggregation_features"]],
            dict_config=DisaggregationModelTrainer.filter_config_by(
                self.config,
                disaggregation_model
            )
        )
        return model



from forecastout.forecast_models.abstract_model import ForecastModel
from prophet import Prophet
import pandas as pd


class ProphetModel(ForecastModel):
    def __init__(self,
                 series_train_dates: pd.Series,
                 *args,
                 **kwargs
                 ):
        """
        This implementation a prophet model
        """
        super(ProphetModel, self).__init__(*args, **kwargs)
        # -- Prophet
        # -- Constructor
        df_train_y_prophet = pd.concat(
            [series_train_dates, self.df_train_y],
            axis=1)
        df_train_y_prophet.columns = ['ds', 'y']
        df_train_y_prophet['ds'] = pd.to_datetime(df_train_y_prophet['ds'])
        df_train_y_prophet['ds'] = (
            df_train_y_prophet['ds']
            .dt
            .tz_localize(None)
        )
        prophet = Prophet()
        self.prophet_model = prophet.fit(df_train_y_prophet)

    def do_forecast(self, list_dates: list) -> pd.DataFrame:
        future = self.prophet_model.make_future_dataframe(
            periods=len(list_dates),
            freq="MS",
            include_history=False
        )
        prophet_prediction = (
            self.prophet_model
            .predict(future)[['ds', 'yhat', 'yhat_lower', 'yhat_upper']]
        )
        prophet_prediction.columns = ['date', 'forecast', 'forecast_lower',
                                      'forecast_upper']
        prophet_prediction['model'] = 'prophet'
        prophet_prediction['date'] = list_dates
        # -- Ensure date
        prophet_prediction['date'] = pd.to_datetime(prophet_prediction['date'])
        return prophet_prediction

    def get_feature_importance(self):
        return None



import pandas as pd


def get_daily_shares(
        df_daily_prediction: pd.DataFrame,
        sum_aggregation: bool
) -> pd.DataFrame:
    df_daily_prediction['year_month'] = (
        df_daily_prediction["date"].astype(str).str[0:7]
    )
    df_daily_prediction['monthly_totals'] = (
        df_daily_prediction
        .groupby(["year_month"])["forecast"].transform(sum)
    )
    if sum_aggregation:
        df_daily_prediction["daily_share"] = (
                df_daily_prediction["forecast"] /
                df_daily_prediction["monthly_totals"]
        )
    else:
        df_daily_prediction["days_month"] = (
            df_daily_prediction
            .groupby(["year_month"])["date"].transform('count')
        )
        df_daily_prediction["daily_share"] = (
                df_daily_prediction["days_month"] *
                df_daily_prediction["forecast"] /
                df_daily_prediction['monthly_totals']
        )
    df_daily_shares = df_daily_prediction[["date", "daily_share"]]
    return df_daily_shares



import pandas as pd


class AnomalyDetector:
    """
    This class determines if there is an anomaly in the forecast
    """

    def __init__(self,
                 df_forecast: pd.DataFrame,
                 df_actual: pd.DataFrame,
                 horizon: int
                 ):
        # -- Main data
        self.df_forecast = df_forecast
        self.df_actual = df_actual
        self.horizon = horizon
        self.columns_id = []
        # -- Ensure order
        self.df_forecast.sort_values(["date"], inplace=True)
        self.df_actual.sort_values(["date"], inplace=True)
        # -- Initialize anomaly 0
        self.anomaly = 0
        # -- Ensure order
        self.__get_explosive_trends()
        self.__get_non_seasonals()
        self.__get_negatives()

    def __get_explosive_trends(self):
        """
        Check trend
        """
        df_forecast = self.df_forecast.copy()
        df_actual = self.df_actual.copy()
        # -- Check all forecast period.
        forecast_allperiod = df_forecast['forecast'].sum()
        # -- Check all actual periods.
        df_actual_allperiod = df_actual.tail(self.horizon * 2)
        actual_period1 = \
            df_actual_allperiod.head(self.horizon)['value'].sum()
        actual_period2 = \
            df_actual_allperiod.tail(self.horizon)['value'].sum()
        # -- Forecast is  < abs( than period1 - period2 dev)*2
        dev_12 = abs(actual_period2 - actual_period1) / actual_period1
        dev_forecast2 = (
                abs(forecast_allperiod - actual_period2) / actual_period2
        )
        if dev_forecast2 > dev_12 * 2:
            self.anomaly = 1

    def __get_non_seasonals(self):
        """
        Check seasonality
        """
        df_forecast = self.df_forecast.copy()
        df_actual = self.df_actual.copy()
        # -- forecast manipulation
        df_forecast = df_forecast.head(
            self.horizon if self.horizon <= 12 else 12
        )
        df_forecast['month'] = df_forecast["date"].dt.month
        df_forecast = df_forecast[["month", "forecast"]].copy()
        # -- actual manipulation
        df_actual = (
            df_actual.tail(self.horizon if self.horizon <= 12 else 12)
        )
        df_actual['month'] = df_actual["date"].dt.month
        df_actual = df_actual[["month", "value"]].copy()
        df_allperiod = df_forecast.merge(
            df_actual,
            on=["month"],
            how='left'
        )
        df_allperiod_corr = df_allperiod[['forecast', 'value']].corr()
        if df_allperiod_corr.iloc[0, 1] < 0.5:
            self.anomaly = 1

    def __get_negatives(self):
        """
        Check negative forecast
        """
        df_forecast = self.df_forecast.copy()
        if len(df_forecast.loc[df_forecast['forecast'] <= 0]) > 0:
            self.anomaly = 1



from forecastout.forecast_models.abstract_model import ForecastModel
from statsforecast.models import HoltWinters
from statsforecast import StatsForecast
import pandas as pd
import numpy as np


class HoltWintersModel(ForecastModel):
    def __init__(self,
                 series_train_dates: pd.Series,
                 *args,
                 **kwargs
                 ):
        """
        This implementation returns the corresponding `ETS` model with additive
        (A) or multiplicative (M) errors (so either 'AAA' or 'MAM').
        """
        super(HoltWintersModel, self).__init__(*args, **kwargs)
        # -- Constructor
        # -- Initialize Holtwinters Exponential Smoothing with additive and
        # multiplicative methods
        self.holtwinters_model = StatsForecast(models=[
            HoltWinters(
                season_length=self.dict_config['freq'],
                error_type='A',
                alias='HW_A'),
            HoltWinters(
                season_length=self.dict_config['freq'],
                error_type='M',
                alias='HW_M')],
            freq='MS',
            n_jobs=-1
        )
        # -- Transform data
        df_train_y_holtwinters = pd.concat(
            [series_train_dates, self.df_train_y],
            axis=1
        )
        df_train_y_holtwinters.columns = ['ds', 'y']
        df_train_y_holtwinters['ds'] = (
            pd.to_datetime(df_train_y_holtwinters['ds'])
        )
        df_train_y_holtwinters['ds'] = (
            df_train_y_holtwinters['ds'].dt.tz_localize(None)
        )
        df_train_y_holtwinters['unique_id'] = 'id'
        # -- Fit forecast_models
        self.holtwinters_model.fit(df_train_y_holtwinters)

    def do_forecast(self, list_dates: list) -> pd.DataFrame:
        # -- Predict
        holtwinters_prediction = self.holtwinters_model.predict(
            h=len(list_dates),
            level=[self.dict_config['alpha_intervals']]
        ).reset_index()
        # -- Transform data for correct output
        value_vars = [
            col for col in holtwinters_prediction.columns
            if 'HW' in col
        ]
        holtwinters_prediction_melted = pd.melt(
            holtwinters_prediction,
            id_vars=["unique_id", "ds"],
            value_vars=value_vars,
            var_name='model_var',
            value_name='value'
        )
        holtwinters_prediction_melted['model'] = (
            holtwinters_prediction_melted['model_var']
            .str.split('-', expand=True)[0]
        )
        holtwinters_prediction_melted['model_metric'] = (
            np.where((
                         holtwinters_prediction_melted['model_var']
                         .str
                         .contains('lo')
                     ) == True,
                     'forecast_lower',
                     np.where((
                                  holtwinters_prediction_melted['model_var']
                                  .str
                                  .contains('hi')
                              ) == True,
                              'forecast_upper',
                              'forecast')
                     )
        )
        holtwinters_prediction = holtwinters_prediction_melted.pivot(
            index=['ds', 'model'],
            columns="model_metric",
            values='value'
        ).reset_index()
        holtwinters_prediction.columns.name = None
        holtwinters_prediction.rename(columns={'ds': 'date'}, inplace=True)
        # -- Ensure date
        holtwinters_prediction['date'] = (
            pd.to_datetime(holtwinters_prediction['date'])
        )
        return holtwinters_prediction

    def get_feature_importance(self):
        return None



import pandas as pd
import numpy as np
import holidays
from datetime import date


class FeatureCreator:

    def __init__(self, df: pd.DataFrame):
        # -- Main data
        self.df = df

    def get_lags(self, features: list, number_lags: int):
        for feature in features:
            for number_lag in range(1, (number_lags+1)):
                column_lag = feature + '_lag_' + str(number_lag)
                self.df[column_lag] = (
                    self.df[feature].shift(number_lag)
                )

    def get_moving_average(self, features: list, windows: list):
        for feature in features:
            for window in windows:
                column_moving_average = feature + '_ma_' + str(window)
                self.df[column_moving_average] = (
                    self.df[feature+'_lag_1']
                        .transform(lambda x: x.rolling(window, center=False)
                                   .mean()
                                   )
                )

    def get_day_of_week(self):
        self.df['day_of_week'] = self.df["date"].dt.dayofweek.astype(np.int64)

    def get_month(self):
        self.df['month'] = self.df["date"].dt.month.astype(np.int64)

    def get_year(self):
        self.df['year'] = self.df["date"].dt.year.astype(np.int64)

    def get_day(self):
        self.df['day'] = self.df["date"].dt.day.astype(np.int64)

    def get_holidays(self):
        # -- Add Spain Holidays
        spain_holidays = holidays.Spain()
        year_list = [i for i in range(2010, 2050)]
        # This is done because of a bug in the package holiday. Weird, yes.
        [date(year, 1, 2) in spain_holidays for year in year_list]
        df_holidays = pd.DataFrame.from_dict(
            spain_holidays,
            orient='index'
        ).reset_index()
        df_holidays["date"] = pd.to_datetime(df_holidays['index'])
        df_holidays['national_holiday'] = 1
        df_holidays.drop([0, 'index'], axis=1, inplace=True)
        # -- Merge holidays with main df
        self.df = self.df.merge(
            df_holidays,
            on="date",
            how='left'
        )
        self.df['national_holiday'] = self.df['national_holiday'].fillna(0)
        # -- Add weekends
        self.df['national_holiday_or_weekend'] = (
            np.where(
                (self.df['national_holiday'] == 1)
                | (self.df['day_of_week'].isin([5, 6])
                   ),
                1,
                0
            )
        )
        # -- Add long weekends
        self.df['puente_1'] = self.df['national_holiday_or_weekend'].shift(1)
        self.df['puente_minus1'] = (
            self.df['national_holiday_or_weekend'].shift(-1)
        )
        self.df['long_weekends'] = (
            np.where((self.df['national_holiday_or_weekend'] == 0)
                     & (self.df['puente_1'] == 1)
                     & (self.df['puente_minus1'] == 1),
                     1,
                     self.df['national_holiday_or_weekend'])
        )
        self.df.drop(["puente_1", "puente_minus1"], axis=1, inplace=True)



from typing import List


def update_config(
        config: dict,
        horizon: int = None,
        months_to_backtest: int = None,
        models_to_use: List[str] = None,
        average_top_models_number: int = None
) -> dict:
    if horizon is not None:
        config["horizon"] = horizon
    if months_to_backtest is not None:
        config["months_to_backtest"] = months_to_backtest
    if models_to_use is not None:
        config["models_to_use"] = models_to_use
    if average_top_models_number is not None:
        config["ensemble"]["average_top_models"]["n_tops"] = \
            average_top_models_number
    if (average_top_models_number is not None) and\
            (average_top_models_number > len(models_to_use)):
        config["ensemble"]["average_top_models"]["n_tops"] = \
            len(models_to_use)
    return config



# -- Packages
import pandas as pd
from dateutil.relativedelta import relativedelta
import calendar
from functools import reduce


class DataHandler:
    """
    This class treats the necessary data to perform the Forecast
    """

    def __init__(self, df: pd.DataFrame, horizon: int, sum_aggregation: bool):
        # -- Main data and parameters
        self.df = df
        self.horizon = horizon
        self.sum_aggregation = sum_aggregation
        self.granularity_of_df = 'monthly'
        # -- Methods to compute granularity
        self.__ensure_date()
        self.__ensure_value()
        self.__detect_data_granularity()
        df_daily, df_monthly = self.__create_granular_dfs()
        # -- Clean df_monthly
        df_monthly = self.__expand_dates(df_monthly, freq='MS')
        df_monthly = self.__nan_to_zero(df_monthly)
        df_monthly = self.__get_order(df_monthly)
        self.df_monthly = df_monthly.copy()
        if self.granularity_of_df == 'daily':
            # -- Clean df_daily
            df_daily = self.__expand_dates(df_daily, freq='D')
            df_daily = self.__nan_to_zero(df_daily)
            df_daily = self.__get_order(df_daily)
            self.df_daily = df_daily.copy()
        else:
            self.df_daily = pd.DataFrame()

    def __ensure_date(self):
        self.df['date'] = pd.to_datetime(self.df['date'])

    def __ensure_value(self):
        self.df['value'] = self.df['value'].astype(float)

    def __detect_data_granularity(self):
        if any(
                self.df.groupby(
                    self.df['date'].astype(str).str[0:7]
                )['date'].count() > 1):
            self.granularity_of_df = "daily"

    def __create_granular_dfs(self) -> (pd.DataFrame, pd.DataFrame):
        df = self.df.copy()
        if self.granularity_of_df == 'monthly':
            df_daily = None
            df_monthly = df.copy()
        else:
            # -- compute relevant dates to clean dataframes
            last_data_date = df['date'].max()
            last_data_day = last_data_date.day
            max_month_day = reduce(
                lambda x, y: y[1],
                [calendar.monthrange(
                    last_data_date.year,
                    last_data_date.month)],
                None
            )
            # -- clean daily dataframe
            if last_data_day < max_month_day:
                first_month_day = pd.to_datetime(
                    str(last_data_date)[0:7] + '-01'
                )
                df_daily = df[df["date"] < first_month_day].copy()
            else:
                df_daily = df.copy()
            # -- clean monthly dataframe
            if self.sum_aggregation:
                df_monthly = (
                    (
                        df_daily
                        .groupby(df['date'].astype(str).str[0:7])['value']
                        .sum()
                        .reset_index()
                    )
                )
            else:
                df_monthly = (
                    (
                        df_daily
                        .groupby(df['date'].astype(str).str[0:7])['value']
                        .mean()
                        .reset_index()
                    )
                )
            df_monthly['date'] = pd.to_datetime(df_monthly['date'] + '-01')
            df_monthly['value'] = df_monthly['value'].astype(float)
        return df_daily, df_monthly

    def __expand_dates(
            self,
            df_granular: pd.DataFrame,
            freq: str
    ) -> pd.DataFrame():
        # -- Get first day of month and max date to disaggregate
        min_date = df_granular["date"].max().replace(day=1)
        if self.granularity_of_df == 'monthly':
            max_date = min_date + relativedelta(months=self.horizon)
        else:
            max_date = (
                    min_date
                    + relativedelta(months=self.horizon+1)
                    + relativedelta(days=-1)
            )
        # -- dates df
        dates = pd.date_range(
            start=df_granular["date"].min(),
            end=max_date,
            freq=freq
        ).tolist()
        df_dates = pd.DataFrame(
            range(len(dates)),
            dates
        ).reset_index()
        df_dates.columns = ["date", 'index']
        # -- Add values
        df_granular = df_dates.merge(
            df_granular,
            on='date',
            how='left'
        )
        df_granular = df_granular.drop('index', axis=1)
        return df_granular

    @staticmethod
    def __get_order(df_granular: pd.DataFrame) -> pd.DataFrame:
        df_granular.sort_values("date", inplace=True)
        return df_granular

    @staticmethod
    def __nan_to_zero(df_granular: pd.DataFrame) -> pd.DataFrame:
        max_actual = pd.to_datetime(df_granular.dropna()['date'].max())
        df_granular.loc[
            (df_granular["value"].isnull())
            & (df_granular["date"] <= max_actual),
            "value"
        ] = 0
        return df_granular



import pandas as pd


def remake_monthly_forecast_current_month(
        df_daily_forecast: pd.DataFrame,
        df_actuals: pd.DataFrame
) -> pd.DataFrame:
    # -- Get actuals of unfinished month
    first_forecast_day = df_daily_forecast['date'].min()
    first_actuals_day = pd.to_datetime(
        str(first_forecast_day)[0:7] + '-01'
    )
    df_actuals = df_actuals.loc[
        (df_actuals['date'] >= first_actuals_day) &
        (df_actuals['date'] < first_forecast_day)
        ].copy()
    df_actuals['forecast'] = df_actuals['value'].copy()
    df_actuals['forecast_lower'] = df_actuals['value'].copy()
    df_actuals['forecast_upper'] = df_actuals['value'].copy()
    df_actuals['model'] = df_daily_forecast['model'].unique()[0]
    df_actuals.drop('value', axis=1, inplace=True)
    # -- Recompute monthly forecast
    df_daily_forecast = pd.concat(
        [df_actuals, df_daily_forecast], axis=0
    )
    df_daily_forecast["date"] = (
        df_daily_forecast["date"].astype(str).str[0:7]
    )
    df_monthly_forecast = (
        df_daily_forecast
        .groupby(["date", "model"])
        [["forecast", "forecast_lower", "forecast_upper"]]
        .sum()
        .reset_index()
    )
    df_monthly_forecast['date'] = pd.to_datetime(
        df_monthly_forecast['date'] + '-01'
    )
    column_order = [
        "forecast",
        "forecast_lower",
        "forecast_upper",
        "date",
        "model"
    ]
    df_monthly_forecast = df_monthly_forecast[column_order].copy()
    return df_monthly_forecast



import pandas as pd
from forecastout.forecast_engine.anomaly_detector import AnomalyDetector
from forecastout.forecast_engine.model_trainer import ModelTrainer
from forecastout.forecast_engine.model_predictor import ModelPredictor


class AnomalyFixer:
    def __init__(
            self,
            df_forecast: pd.DataFrame,
            df_actual: pd.DataFrame,
            config: dict
    ):
        self.df_forecast = df_forecast
        self.df_actual = df_actual
        self.config = config
        # -- Anomalies
        anomaly_detector = AnomalyDetector(
            df_forecast=df_forecast.copy(),
            df_actual=df_actual[df_actual['test'] == 0].copy(),
            horizon=self.config['horizon']
        )
        self.anomaly = anomaly_detector.anomaly
        self.__apply_recursive_naiveseasonal()

    def __apply_recursive_naiveseasonal(self):
        i = 0
        while i < 100 and self.anomaly > 0:
            # -- Forecast with models_to_fix
            model_trainer = ModelTrainer(
                model_names=self.config["models_to_fix"],
                df_train=self.df_actual[self.df_actual['test'] == 0].copy(),
                config=self.config
            )
            df_models_trained = model_trainer.train()
            model_predictor = ModelPredictor(
                df_models_trained=df_models_trained,
                df=self.df_actual[self.df_actual['test'] == 1].copy(),
                config=self.config,
            )
            df_predictions = model_predictor.predict()
            # -- Correct forecast
            if i > 0:
                df_predictions['model'] = ''
            model_sep = '' if i > 0 else '/'
            df_forecast_corrected = pd.concat(
                [self.df_forecast, df_predictions]
            )
            self.df_forecast = df_forecast_corrected.groupby(["date"]).agg({
                'forecast': 'mean',
                'forecast_lower': 'mean',
                'forecast_upper': 'mean',
                'model': lambda x: model_sep.join(x)}).reset_index()
            anomaly_detector = AnomalyDetector(
                df_forecast=self.df_forecast.copy(),
                df_actual=(
                    self.df_actual[self.df_actual['test'] == 0].copy()
                ),
                horizon=self.config['horizon']
            )
            # -- Check repetition
            self.anomaly = anomaly_detector.anomaly
            i += 1
        



# Main forecasting forecast_models

# -- Python main packages
from forecastout.data_engine.train_test_split import TrainTestSplit
from forecastout.forecast_engine.model_trainer import ModelTrainer
from forecastout.forecast_engine.model_predictor import ModelPredictor
from typing import List
import pandas as pd


class Backtester:
    def __init__(self, df: pd.DataFrame, model_names: List[str], config: dict):
        self.df = df
        self.model_names = model_names
        self.config = config
        self.df_predictions = pd.DataFrame()
        self.df_models_ranked = pd.DataFrame()
        # -- start process
        self.__make_forecast_for_validation()
        self.__rank_models()

    def __make_forecast_for_validation(self):
        # --  BT. Train/Test Split
        df = TrainTestSplit.split_for_ts_backtest(
            df=self.df,
            current_month=self.df.loc[~self.df['value'].isna(), 'date'].max(),
            months_to_backtest=self.config["months_to_backtest"])
        # --  BT. Train Models
        model_trainer = ModelTrainer(
            model_names=self.config["models_to_use"],
            df_train=df[df['test'] == 0].copy(),
            config=self.config
        )
        df_models_trained = model_trainer.train()
        # -- BT. Predict Models
        model_predictor = ModelPredictor(
            df_models_trained=df_models_trained,
            df=df[df['test'] == 1].copy(),
            config=self.config,
        )
        df_predictions = model_predictor.predict()
        # -- BT. Merge actual and predictions
        df_predictions = df_predictions.merge(
            df.loc[df['test'] == 1, ["date", "value"]].copy(),
            on="date",
            how='left'
        )
        self.df_predictions = (
            df_predictions[["date", "model", "forecast", "value"]]
            .copy()
        )

    def __rank_models(self):
        # -- BT. Rank Models
        df_predictions = self.df_predictions.copy()
        df_predictions['abs_error'] = (
                abs((df_predictions['value'] - df_predictions['forecast']))
                / df_predictions['value']
        )
        df_ranking = (
            df_predictions
            .groupby(["model"])['abs_error']
            .mean()
            .reset_index()
        )
        df_ranking.rename(
            columns={'abs_error': 'mean_abs_error'},
            inplace=True
        )
        df_ranking.sort_values(["mean_abs_error"], inplace=True)
        df_ranking['ranking'] = 1
        df_ranking['ranking'] = df_ranking['ranking'].cumsum()
        self.df_models_ranked = df_ranking.copy()

    def return_predictions(self) -> pd.DataFrame:
        return self.df_predictions

    def return_models_ranked(self) -> pd.DataFrame:
        return self.df_models_ranked



from forecastout.disaggregation_engine.disaggregation_model_trainer \
    import DisaggregationModelTrainer
from forecastout.disaggregation_engine.disaggregation_model_predictor \
    import DisaggregationModelPredictor
from forecastout.disaggregation_engine.get_daily_shares import \
    get_daily_shares
from forecastout.disaggregation_engine\
    .disaggregate_monthly_to_daily_forecast \
    import disaggregate_monthly_to_daily_forecast
from forecastout.disaggregation_engine\
    .remake_monthly_forecast_current_month \
    import remake_monthly_forecast_current_month



import pandas as pd
from typing import List


class FeatureEncoder:

    @staticmethod
    def numerical_encode(
            df: pd.DataFrame,
            encode_feats: List[str],
            encode_value: str) -> pd.DataFrame:

        for encode_feat in encode_feats:
            df_encoded = (
                df.groupby(encode_feat)[encode_value].mean().reset_index()
            )
            df_encoded.rename(
                columns={encode_value: encode_feat+'_num_enc'},
                inplace=True
            )
            df = df.merge(
                df_encoded,
                on=encode_feat, how='left')
        return df
